#
# FtpCube
# Copyright (C) 2001 Michael Gilfix
#
# This file is part of FtpCube.
#
# You should have received a file COPYING containing license terms
# along with this program; if not, write to Michael Gilfix
# (mgilfix@eecs.tufts.edu) for a copy.
#
# This version of FtpCube is open source; you can redistribute it and/or
# modify it under the terms listed in the file COPYING.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#

import gtk, GtkExtra
import sys, os
import string # For backwards compatibility with 1.5.x
import re
import main

from intl import _

# For compatibility with win32 pygtk
if sys.platform == "win32":
	import GTKconst
	GTK = GTKconst
else:
	import GTK

class ConnectWin:

	"""Connection window object

	Brings up a new window displaying connection information and "Ok"/"Cancel" options.
	The tabs refer to different connection parameters and the user inputs the desired
	host to connect to and whether to use login/password or anonymous access. Other
	connection parameters such as number of retry attempts and idle timeout periods can
	be set in the connection window.

	The connection window is created with reference to a callback function for
	actually initiating the connection provided the user clicks the "Ok" button. A hash
	of default options is also provided, indicating which values to use for seeding
	the various input boxes."""

	def __init__ (self, callback, default_opts):
		"""Creates and displays the connection window object. 'callback' is a reference
		to a callback function that is executed when the "Ok" button is clicked. The
		definition of 'callback' is:

			def callback (connectwin):
				# Where 'connectwin' is a reference to the calling ConnectWin instance

		'default_opts' is a hash containing the default options to seed the input
		boxes with."""

		self.callback = callback

		self.dialog = gtk.GtkDialog ()
		self.dialog.set_title (_("ftpcube - Connect"))
		self.dialog.set_policy (gtk.FALSE, gtk.FALSE, gtk.TRUE)
		self.dialog.set_position (gtk.WIN_POS_CENTER)
		icon_pic, icon_mask = gtk.create_pixmap_from_xpm (self.dialog, None, os.path.join (main.ICONS_PREFIX, "ftpcube.xpm"))
		self.dialog.set_icon (gtk.GtkPixmap (icon_pic, icon_mask))

		# Make ourselves modal
		gtk.grab_add (self.dialog)

		box = gtk.GtkHBox (spacing=75)
		box.show ()
		self.dialog.action_area.add (box, padding=95)

		ok_button = gtk.GtkButton (_("Ok"))
		ok_button.set_usize (50, 25)
		ok_button.connect ("clicked", self.ok)
		box.pack_start (ok_button, expand=gtk.FALSE)
		ok_button.show ()

		button = gtk.GtkButton (_("Cancel"))
		button.set_usize (50, 25)
		button.connect ("clicked", self.cancel)
		box.pack_start (button, expand=gtk.FALSE)
		button.show ()

		self.notebook = gtk.GtkNotebook ()
		self.notebook.set_tab_pos (gtk.POS_TOP)
		self.dialog.vbox.add (self.notebook)
		self.notebook.show ()

		# Create the tabs and add them and show them
		self.login_tab = LoginTab (self, opts=default_opts)
		label = gtk.GtkLabel (_("Login Info"))
		label.set_padding (2, 2)
		label.show ()
		self.notebook.append_page (self.login_tab.get_page (), label)

		self.advanced_tab = AdvancedTab (self, opts=default_opts)
		label = gtk.GtkLabel (_("Advanced"))
		label.set_padding (2, 2)
		label.show ()
		self.notebook.append_page (self.advanced_tab.get_page (), label)

		self.settings_tab = SettingsTab (self, opts=default_opts)
		label = gtk.GtkLabel (_("Settings"))
		label.set_padding (2, 2)
		label.show ()
		self.notebook.append_page (self.settings_tab.get_page (), label)

		self.dialog.connect ("key_press_event", self.key_press)
		self.dialog.show ()

	def key_press (self, win, event):

		if hasattr (event, 'string') and event.string == '\r':
			self.ok (None)

	def ok (self, button):
		"""Destroys the connect window and executes the callback."""

		self.cancel (button)
		self.callback (self)

	def cancel (self, button):
		"""Destroys the connect window by hiding it."""

		self.dialog.hide ()
		self.dialog.destroy ()

	def get_options (self):
		"""Returns a hash containing a list of all the values of the
		various input boxes, containing the default values and those
		the user modified."""

		options = { }
		options.update (self.login_tab.get_options ())
		options.update (self.advanced_tab.get_options ())
		options.update (self.settings_tab.get_options ())
		return options

class LoginTab:

	"""Connect window login tab

	Displays a tab containing information detailing which host/port to
	connect to and what login/password combination to use. A remote
	directory can also be specified. If the directory box is blank,
	'/' is used."""

	ftp_url_re = r'ftp://(([^:]+)(:([^@]+))?@)?([a-zA-Z\d\.-]+)(:(\d+))?(/.+)?'

	def __init__ (self, connectwin, ok_button=None, opts=None):
		"""Creates an instance of the tab object and displays the tab. 'connectiwin'
		is a reference to the parent connect window object. 'opts' is a reference to
		the defaults hash provided to the parent connect window object."""

		self.connectwin = connectwin
		self.ok_button = ok_button

		self.page = gtk.GtkFrame ()
		self.page.set_border_width (10)
		self.page.set_shadow_type (gtk.SHADOW_ETCHED_OUT)
		self.page.show ()

		table = gtk.GtkTable (6, 1)
		table.set_row_spacings (2)
		table.set_col_spacings (5)
		table.set_homogeneous (gtk.FALSE)
		table.show ()
		self.page.add (table)

		packer = gtk.GtkPacker ()
		label = gtk.GtkLabel (_("Host or URL"))
		label.show ()
		packer.show ()
		packer.add (label, anchor=GTK.ANCHOR_W)
		table.attach (packer, 0, 1, 0, 1, xpadding=12, ypadding=1)

		packer = gtk.GtkPacker ()
		label = gtk.GtkLabel (_("Port"))
		label.show ()
		packer.show ()
		packer.add (label, anchor=GTK.ANCHOR_W)
		table.attach (packer, 1, 2, 0, 1, xpadding=10, ypadding=1)

		self.host = gtk.GtkEntry ()
		if opts is not None and opts['host']:
			self.host.set_text (opts['host'])
		self.host.connect ("focus_out_event", self.convert_hostaddr)
		self.host.show ()
		table.attach (self.host, 0, 1, 1, 2, xoptions=GTK.FILL, xpadding=4, ypadding=1)

		packer = gtk.GtkPacker ()
		self.port = gtk.GtkEntry ()
		self.port.set_usize (50, -1)
		self.port.set_text ("%s" %main.app['port'])
		if opts is not None and opts['port']:
			self.port.set_text ("%s" %opts['port'])
		self.port.show ()
		packer.show ()
		packer.add (self.port, anchor=GTK.ANCHOR_W)
		table.attach (packer, 1, 2, 1, 2, xpadding=4, ypadding=1)

		packer = gtk.GtkPacker ()
		label = gtk.GtkLabel (_("Username"))
		label.show ()
		packer.show ()
		packer.add (label, anchor=GTK.ANCHOR_W)
		table.attach (packer, 0, 1, 2, 3, xpadding=12, ypadding=1)

		packer = gtk.GtkPacker ()
		label = gtk.GtkLabel (_("Password"))
		label.show ()
		packer.show ()
		packer.add (label, anchor=GTK.ANCHOR_W)
		table.attach (packer, 1, 2, 2, 3, xpadding=12, ypadding=1)

		self.username = gtk.GtkEntry ()
		if opts is not None and opts['username']:
			self.username.set_text (opts['username'])
		self.username.connect ("key_press_event", self.entry_update)
		self.username.show ()
		table.attach (self.username, 0, 1, 3, 4, xpadding=4, ypadding=1)

		self.password = gtk.GtkEntry ()
		self.password.set_visibility (gtk.FALSE)
		if opts is not None and opts['password']:
			self.password.set_text (opts['password'])
		self.password.connect ("key_press_event", self.entry_update)
		self.password.show ()
		table.attach (self.password, 1, 2, 3, 4, xpadding=4,  ypadding=1)

		packer = gtk.GtkPacker ()
		label = gtk.GtkLabel (_("Remote Directory"))
		label.show ()
		packer.show ()
		packer.add (label, anchor=GTK.ANCHOR_W)
		table.attach (packer, 0, 2, 4, 5, xpadding=12, ypadding=1)

		self.remote_dir = gtk.GtkEntry ()
		if opts is not None and opts['remotedir']:
			self.remote_dir.set_text (opts['remotedir'])
		self.remote_dir.show ()
		table.attach (self.remote_dir, 0, 2, 5, 6, xpadding=4, ypadding=1)

		box = gtk.GtkHBox ()
		box.show ()
		self.personal_button = gtk.GtkRadioButton (None, _("Personal Login"))
		self.personal_button.connect ("clicked", self.personal_clicked)
		self.personal_button.show ()
		self.anon_button = gtk.GtkRadioButton (self.personal_button, _("Anonymous Login"))
		self.anon_button.connect ("clicked", self.anon_clicked)
		self.anon_button.show ()
		if opts is not None and not opts['username'] == main.app['username']:
			self.personal_button.set_active (gtk.TRUE)
		else:
			self.anon_button.set_active (gtk.TRUE)
		box.pack_start (self.personal_button)
		box.pack_start (self.anon_button)
		table.attach (box, 0, 2, 6, 7, xpadding=4, ypadding=1)

	def convert_hostaddr (self, win, event):
		"""Convert a hostname of the form
		ftp://user@host:domain.net:1234/initial/path and put the data in
		the correct fields."""

		conn = re.match (self.ftp_url_re, self.host.get_text ())
		if conn:
			conninfo = conn.groups ()
			#('user:pass@','user',':pass','pass','host.com',':123','1231','/dir')
			if conninfo[1]:
				if conninfo[1] == main.app['host']:
					self.anon_button.set_active (gtk.TRUE)
				else:
					self.personal_button.set_active (gtk.TRUE)
				self.username.set_text (conninfo[1])
				if conninfo[3]:
					self.password.set_text (conninfo[3])
			else:
				self.anon_clicked (None)
			self.host.set_text (conninfo[4])
			if conninfo[6]:
				self.port.set_text (conninfo[6])
			else:
				self.port.set_text (main.app['port'])
			if conninfo[7]:
				self.remote_dir.set_text (conninfo[7])
			else:
				self.remote_dit.set_text ('')

	def entry_update (self, win, event):
		"""Automatically toggles between anonymous and personal radio buttons provided
		that the user begins to enter a new login/password combination."""

		self.personal_button.set_active (gtk.TRUE)

	def get_page (self):
		return self.page

	def anon_clicked (self, button):
		"""If the anonymous button is clicked, the login/password combination are set
		to the defaults of the application, which are usually 'anonymous' and some
		password 'guest@password.org'."""

		self.username.set_text ("%s" %main.app['username'])
		self.password.set_text ("%s" %main.app['password'])

	def personal_clicked (self, button):
		"""If personal is clicked, the login/password combination are blanked."""

		self.username.set_text ('')
		self.password.set_text ('')

	def get_options (self):
		"""Returns a hash containing the information from each of the input boxes."""

		# These are safe to strip and are probably expected to strip
		host = self.host.get_text ()
		host = string.strip (host)
		username = self.username.get_text ()
		username = string.strip (username)

		return {
			'host' : host,
			'port' : int (self.port.get_text ()),
			'username' : username,
			'password' : self.password.get_text (),
			'remotedir' : self.remote_dir.get_text ()
		}

class AdvancedTab:

	"""Connect window advanced tab

	Contains information about the local directory to default to when connecting to
	the host (useful for bookmarks) and possible limits for logins that can affect
	how threading works."""

	def __init__ (self, connectwin, ok_button=None, opts=None):
		"""Creates and displays the advanced tab. 'connectwin' is a reference to
		the calling connect window object. 'opts' is a hash containing the default
		option data passed to the connect window object."""

		self.connectwin = connectwin

		self.page = gtk.GtkFrame ()
		self.page.set_border_width (10)
		self.page.set_shadow_type (gtk.SHADOW_ETCHED_OUT)
		self.page.show ()

		outterpacker = gtk.GtkPacker ()
		outterpacker.show ()
		self.page.add (outterpacker)

		frame = gtk.GtkFrame ()
		frame.set_border_width (10)
		frame.set_shadow_type (gtk.SHADOW_ETCHED_OUT)
		frame.set_usize (350, 85)
		frame.show ()
		outterpacker.add (frame)

		packer = gtk.GtkPacker ()
		packer.show ()
		frame.add (packer)
		label = gtk.GtkLabel (_("Local Directory"))
		label.show ()
		packer.add (label, side=GTK.SIDE_TOP, anchor=GTK.ANCHOR_W, pad_x=75, pad_y=12)
		box = gtk.GtkHBox ()
		box.show ()
		packer.add (box)
		self.localdir = gtk.GtkEntry ()
		if opts is not None and opts['localdir']:
			self.localdir.set_text (opts['localdir'])
		self.localdir.set_usize (200, -1)
		self.localdir.show ()
		box.pack_start (self.localdir, expand=GTK.EXPAND)
		button = gtk.GtkButton ("...")
		button.connect ("clicked", self.browse)
		button.set_usize (25, 25)
		button.show ()
		box.pack_start (button, expand=GTK.EXPAND, padding=25)

		frame = gtk.GtkFrame ()
		frame.set_border_width (10)
		frame.set_usize (350, 85)
		frame.set_shadow_type (gtk.SHADOW_ETCHED_OUT)
		frame.show ()
		outterpacker.add (frame)

		packer = gtk.GtkPacker ()
		packer.show ()
		frame.add (packer)
		label = gtk.GtkLabel (_("Number of logins allowed:"))
		label.show ()
		packer.add (label, side=GTK.SIDE_TOP, anchor=GTK.ANCHOR_NW, pad_x=55, pad_y=10)
		box = gtk.GtkHBox ()
		box.show ()
		self.unlimited_button = gtk.GtkRadioButton (None, _("Unlimited"))
		self.unlimited_button.show ()
		box.pack_start (self.unlimited_button, padding=4)
		self.unlimited_button.set_active (gtk.TRUE)
		self.login_button = gtk.GtkRadioButton (self.unlimited_button, _("Login only once"))
		self.login_button.show ()
		box.pack_start (self.login_button, padding=4)
		self.limited_button = gtk.GtkRadioButton (self.unlimited_button, _("Limited"))
		self.limited_button.show ()
		box.pack_start (self.limited_button, padding=4)
		self.limited_box = gtk.GtkEntry ()
		if opts is not None and opts['limit']:
			limit = opts['limit']
			if limit == 0:
				self.unlimited_button.set_active (gtk.TRUE)
			elif limit == 1:
				self.login_button.set_active (gtk.TRUE)
			else:
				self.unlimited_button.set_active (gtk.TRUE)
				self.limited_box.set_text (limit)
		self.limited_box.set_usize (25, 25)
		self.limited_box.show ()
		box.pack_start (self.limited_box, expand=GTK.EXPAND, padding=4)
		packer.add (box, anchor=GTK.ANCHOR_W, pad_x=15)

	def get_page (self):
		return self.page

	def browse (self, button):
		"""Brings up a file selection window where the user can select the
		local path that will appear in the "Local Directory" input box."""

		# Release our modal grabbing of the window
		gtk.grab_remove (self.connectwin.dialog)

		file_win = gtk.GtkFileSelection (_("ftpcube - Browse"))
		file_win.show ()

		# Create the handlers for browse events
		def win_close (button, win=file_win, self=self):
			win.hide ()
			win.destroy ()
			gtk.grab_add (self.connectwin.dialog)
		def browse_selected (button, fs=file_win, self=self):
			self.localdir.set_text (fs.get_filename ())
			fs.hide ()
			fs.destroy ()
			gtk.grab_add (self.connectwin.dialog)
		file_win.ok_button.connect("clicked", browse_selected)
		file_win.cancel_button.connect("clicked", win_close)

	def get_options (self):
		"""Returns a hash containing the information in the tab's input boxes."""

		limit = None
		if self.login_button.get_active ():
			limit = 1
		elif self.limited_button.get_active ():
			text = self.limited_box.get_text ()
			if len (text) > 0:
				limit = int (text)

		return {
			'limit' : limit,
			'localdir' : self.localdir.get_text ()
		}

class SettingsTab:

	"""Connect window settings tab

	Displays a tab allowing the user to set various connection parameters, including
	number of retires, connection/idle timeouts, and delay between login retries. The
	user can also indicate whether a proxy server is being used."""

	def __init__ (self, connectwin, ok_button=None, opts=None):
		"""Creates and displays the Settings tab. 'connectwin' is a reference to the
		calling connect window object. 'opts' is a reference to the default hash passed
		to the connect window."""

		self.connectwin = connectwin

		self.page = gtk.GtkFrame ()
		self.page.set_border_width (10)
		self.page.set_shadow_type (gtk.SHADOW_ETCHED_OUT)
		self.page.show ()

		outterpacker = gtk.GtkPacker ()
		outterpacker.show ()
		self.page.add (outterpacker)

		frame = gtk.GtkFrame ()
		frame.set_border_width (10)
		frame.set_shadow_type (gtk.SHADOW_ETCHED_OUT)
		frame.set_usize (350, 130)
		frame.show ()
		outterpacker.add (frame)
 
		table = gtk.GtkTable (2, 4)
		table.set_col_spacing (0, 15)
		table.show ()
		frame.add (table)
		label = gtk.GtkLabel (_("Connection Retries"))
		label.show ()
		table.attach (label, 0, 1, 0, 1)
		label = gtk.GtkLabel (_("Idle Timeout (sec)"))
		label.show ()
		table.attach (label, 1, 2, 0, 1)
		self.retries_box = gtk.GtkEntry ()
		self.retries_box.set_usize (100, -1)
		if opts is not None and opts['retries']:
			self.retries_box.set_text ("%s" %opts['retries'])
		self.retries_box.show ()
		table.attach (self.retries_box, 0, 1, 1, 2, xoptions=GTK.EXPAND)
		self.idle_box = gtk.GtkEntry ()
		self.idle_box.set_usize (100, -1)
		if opts is not None and opts['main_idle']:
			self.idle_box.set_text ("%s" %opts['main_idle'])
		self.idle_box.show ()
		table.attach (self.idle_box, 1, 2, 1, 2, xoptions=GTK.EXPAND)
		label = gtk.GtkLabel (_("Connect timeout (sec)"))
		label.show ()
		table.attach (label, 0, 1, 2, 3)
		label = gtk.GtkLabel (_("Login retry delay (sec)"))
		label.show ()
		table.attach (label, 1, 2, 2, 3)
		self.timeout_box = gtk.GtkEntry ()
		self.timeout_box.set_usize (100, -1)
		if opts is not None and opts['timeout']:
			self.timeout_box.set_text ("%s" %opts['timeout'])
		self.timeout_box.show ()
		table.attach (self.timeout_box, 0, 1, 3, 4, xoptions=GTK.EXPAND)
		self.delay_box = gtk.GtkEntry ()
		self.delay_box.set_usize (100, -1)
		if opts is not None and opts['delay']:
			self.delay_box.set_text ("%s" %opts['delay'])
		self.delay_box.show ()
		table.attach (self.delay_box, 1, 2, 3, 4, xoptions=GTK.EXPAND)

		frame = gtk.GtkFrame ()
		frame.set_border_width (10)
		frame.set_shadow_type (gtk.SHADOW_ETCHED_OUT)
		frame.set_usize (350, 65)
		frame.show ()
		outterpacker.add (frame)

		box = gtk.GtkHBox ()
		box.show ()
		frame.add (box)
		self.proxy_check = gtk.GtkCheckButton (_("Use proxy server"))
		if opts is not None and opts['proxy']:
			self.proxy_check.set_active (gtk.TRUE)
		self.proxy_check.show ()
		box.pack_start (self.proxy_check, padding=100)

	def get_page (self):
		return self.page

	def get_options (self):
		"""Returns a hash containing the information in the tab's input boxes."""

		return {
			'retries' : int (self.retries_box.get_text ()),
			'main_idle' : int (self.idle_box.get_text ()),
			'timeout' : int (self.timeout_box.get_text ()),
			'delay' : int (self.delay_box.get_text ()),
			'proxy' : self.proxy_check.get_active ()
		}

class QuickConnectWin:

	"""Quick connection window

	This object brings up a pop-up window which only asks the user which host to
	connect to and does so assuming the rest of the application default parameters
	for connecting to an FTP server."""

	def __init__ (self, callback, opts):
		"""Create the quick connect object and display the pop-up. 'callback' is a reference
		to the function to execute provided the user has provided a host to connect to and
		clicked "Ok". 'opts' is a reference to the default options to assume when connecting
		(usually the application defaults)."""

		self.callback = callback

		self.opts = { }
		# Iterate through the keys since opts is not a real hash instance
		# but just an instance of Application
		if opts is not None:
			for key in opts.keys ():
				self.opts[key] = opts[key]

		self.host = GtkExtra.input_box (title=_("ftpcube - Quick Connect"), message=_("Remote Host:"))

		if not self.host:
			return
		else:
			self.callback (self)

	def get_options (self):
		"""Returns a hash containing all of the application defaults for connecting as well
		as the specified host."""

		opts = self.opts
		self.host = string.strip (self.host)
		opts['host'] = self.host

		# Perform all conversions on the main.app defaults
		# to integers that require conversion
		if main.app['port']:
			opts['port'] = int (main.app['port'])
		if main.app['retries']:
			opts['retries'] = int (main.app['retries'])
		if main.app['main_idle']:
			opts['main_idle'] = int (main.app['main_idle'])
		if main.app['timeout']:
			opts['timeout'] = int (main.app['timeout'])
		if main.app['delay']:
			opts['delay'] = int (main.app['delay'])

		return opts



#
# FtpCube
# Copyright (C) 2001 Michael Gilfix
#
# This file is part of FtpCube.
#
# You should have received a file COPYING containing license terms
# along with this program; if not, write to Michael Gilfix
# (mgilfix@eecs.tufts.edu) for a copy.
#
# This version of FtpCube is open source; you can redistribute it and/or
# modify it under the terms listed in the file COPYING.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#

import types
import string # For backwards compatibility with 1.5.x

class CacheNode:

	"""A Cache Node

	A container object for a node in the cache tree."""

	def __init__ (self):

		self.name = None
		self.parent = None
		self.data = None
		self.children = { }

class CacheTree:

	"""A Tree Caching object.

	Provides a cache for storing hierachical data, specifically file
	paths. Paths are given by node names separated by a delimiter.
	The CacheTree allows for such operations as node creation,
	searching, and deletion."""

	def __init__ (self, delim='/'):
		"""Creates the cache by initializing the root. 'delim'
		is the character delimiter that separates the hierachical
		path between node names."""

		self.root = CacheNode ()
		self.root.name = delim
		self.delim = delim

	def get_delim (self):
		"""Returns the delimeter."""

		return self.delim

	def __path_to_list (self, path):
		"""Takes the hierarchical 'path' and splits it using the
		delimiter. Blank entries from the split are then filtered.
		"""

		if type (path) == types.StringType:
			names = filter (lambda x: x, string.split (path, self.delim))
		elif type (path) == types.ListType:
			names = path
		else:
			raise TypeError, "called find_node with invalid path type"
		return names

	def find_node (self, path):
		"""Finds the node associated with the 'path'. Returns the node
		if found, and None otherwise."""

		names = self.__path_to_list (path)
		node = self.root

		while names and node is not None:
			node_name = names.pop (0)
			if node.children.has_key (node_name):
				node = node.children[node_name]
			else:
				node = None
		return node

	def get_node_list (self, path):
		"""Returns a list of nodes along the 'path'. If the path is
		invalid, None is returned."""

		names = self.__path_to_list (path)
		node = self.root
		node_list = [ node ]

		while names and node is not None:
			node_name = names.pop (0)
			if node.children.has_key (node_name):
				node = node.children[node_name]
				node_list.append (node)
			else:
				node = None
		return node_list

	def get_root (self):
		"""Returns the root node."""

		return self.root

	def get_children (self, node):
		"""Returns the list of children associated with the root node.
		No order is preserved here."""

		return node.children.keys ()

	def get_child_node (self, node, child):
		"""Returns the object node 'child' who is a valid child of 'node'.
		Otherwise, None is returned."""

		if node.children.has_key (child):
			return node.children[child]
		return None

	def remove_child (self, node, child):
		"""Removes a 'child' from the cache tree with 'node' as its
		parent."""

		if node.children.has_key (child):
			child = node.children[child]
			del node.children[child]
			return child
		return None

	def remove_node (self, path):
		"""Removes the node found at 'path' from the tree."""

		node = self.find_node (path)
		if node.parent is not None:
			del node.parent.children[node.name]
		return node

	def add_node (self, path):
		"""Adds a node whose entry in the hierachical tree corresponds to
		'path'. If sub nodes are missing during the creation, they are
		created and thier children set accordingly."""

		new_node = CacheNode ()
		node = self.root
		names = filter (lambda x: x, string.split (path, self.delim))
		if not names:
			return node
		name = names.pop ()

		while names:
			node_name = names.pop (0)
			if not node.children.has_key (node_name):
				path_node = CacheNode ()
				path_node.name = node_name
				node.children[node_name] = path_node
			node = node.children[node_name]

		new_node.name = name
		new_node.parent = node
		node.children[name] = new_node
		return new_node



#
# FtpCube
# Copyright (C) 2001 Michael Gilfix
#
# This file is part of FtpCube.
#
# You should have received a file COPYING containing license terms
# along with this program; if not, write to Michael Gilfix
# (mgilfix@eecs.tufts.edu) for a copy.
#
# This version of FtpCube is open source; you can redistribute it and/or
# modify it under the terms listed in the file COPYING.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#

# Set the '_' function for i18n translation
try:
	import gettext

	gettext.binddomain ('ftpcube')
	gettext.textdomain ('ftpcube')
	_ = gettext.gettext
except Exception, strerror:
	_ = lambda x: x



#
# FtpCube
# Copyright (C) 2001 Michael Gilfix
#
# This file is part of FtpCube.
#
# You should have received a file COPYING containing license terms
# along with this program; if not, write to Michael Gilfix
# (mgilfix@eecs.tufts.edu) for a copy.
#
# This version of FtpCube is open source; you can redistribute it and/or
# modify it under the terms listed in the file COPYING.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#

import gtk, GtkExtra
import string # For backwards compatibility with 1.5.x
import time, sys, os
import main, app, ftp, logging

from intl import _

# For compatibility with win32 pygtk
if sys.platform == "win32":
	import GTKconst
	GTK = GTKconst
else:
	import GTK

class FtpThread:

	def __init__ (self, connect_opts, name=None, silence=0):

		# Instance variables
		self.progress = (0, 0)
		self.start_time = 0.0
		self.idle = 0
		self.done = 0
		self.idle_disabled = 0
		self.name = name
		self.silence = silence
		self.opts = connect_opts
		self.logging = 0
		self.logs = {
			'download_log' : None,
			'upload_log'   : None,
		}

		# Create an event box to catch all button presses on the lower widgets
		self.entry_box = gtk.GtkEventBox ()
		self.entry_box.connect ("button_press_event", self.popup_menu)
		self.entry_box.show ()

		# Load the pixmaps for the action pictures
		self.load_pixmaps ()

		vbox = gtk.GtkVBox ()
		vbox.show ()
		self.entry_box.add (vbox)

		self.progress_packer = gtk.GtkPacker ()
		self.progressbar = gtk.GtkProgressBar ()
		self.progressbar.set_usize (200, 10)
		self.progress_packer.add (self.progressbar, options=GTK.FILL, pad_y=4)
		vbox.pack_start (self.progress_packer)

		table = gtk.GtkTable (2, 4)
		table.show ()
		vbox.pack_start (table, expand=gtk.FALSE, fill=gtk.FALSE)

		# Button and label
		self.pixmap = gtk.GtkPixmap (self.connect_pixmap, self.connect_mask)
		self.pixmap.set_usize (20, 20)
		self.pixmap.show ()
		table.attach (self.pixmap, 0, 1, 1, 2, xoptions=GTK.EXPAND, yoptions=GTK.EXPAND, xpadding=2)

		self.label = gtk.GtkLabel ()
		self.label.set_line_wrap (gtk.TRUE)
		self.label.set_usize (200, -1)
		self.label.show ()
		table.attach (self.label, 1, 2, 0, 3, xoptions=GTK.EXPAND, yoptions=GTK.EXPAND, xpadding=2)

		separator = gtk.GtkHSeparator ()
		separator.show ()
		table.attach (separator, 0, 2, 3, 4)

		self.initiate_connect ()
		self.update_local_listing ()

	def load_pixmaps (self):

		self.idle_pixmap, self.idle_mask = gtk.create_pixmap_from_xpm (self.entry_box, None, os.path.join (main.ICONS_PREFIX, "thread_idle.xpm"))
		self.connect_pixmap, self.connect_mask = gtk.create_pixmap_from_xpm (self.entry_box, None, os.path.join (main.ICONS_PREFIX, "thread_connect.xpm"))
		self.listing_pixmap, self.listing_mask = gtk.create_pixmap_from_xpm (self.entry_box, None, os.path.join (main.ICONS_PREFIX, "thread_listing.xpm"))
		self.abort_pixmap, self.abort_mask = gtk.create_pixmap_from_xpm (self.entry_box, None, os.path.join (main.ICONS_PREFIX, "thread_abort.xpm"))
		self.action_pixmap, self.action_mask = gtk.create_pixmap_from_xpm (self.entry_box, None, os.path.join (main.ICONS_PREFIX, "thread_action.xpm"))
		self.download_pixmap, self.download_mask = gtk.create_pixmap_from_xpm (self.entry_box, None, os.path.join (main.ICONS_PREFIX, "thread_download.xpm"))

	def get_thread_name (self):

		if self.name is None:
			return ''
		else:
			return "[%s] " %self.name

	def popup_menu (self, widget, event):

		if event.button == 3:
			opts = [
				(_('/Cancel Download'), None, self.cancel_download),
				(_('/Cancel Thread'), None, self.cancel_thread),
				('/<separator>', None, None),
			]

			if not self.idle_disabled:
				opts.append ((_('/Disable Idle Timeout'), None, self.disable_idle_timeout))
			else:
				opts.append ((_('/Enable Idle Timeout'), None, self.enable_idle_timeout))

			menu = GtkExtra.MenuFactory (GtkExtra.MENU_FACTORY_MENU)
			menu.add_entries (opts)
			menu.popup (None, None, None, event.button, event.time)
			menu.show ()

	def initiate_connect (self):

		self.pixmap.set (self.connect_pixmap, self.connect_mask)
		self.label.set_text (self.get_thread_name () + _("Host: %s\nConnecting...") %(self.opts['host']))

		self.connect = ftp.FTP (
			host=self.opts['host'],
			port=self.opts['port'],
			silence=self.silence
		)

		def handler (msg, self=self):

			gtk.threads_enter ()
			main.app.ftp_error_tell_user (msg)
			gtk.threads_leave ()
			self.done = 1

		self.set_exception_handler (handler)
		self.connect.initiate_connect ()

		# Set ourselves in accordance with the main.app default
		self.set_log_active (main.app['logging'])

		self.connect.get_welcome_msg ()
		self.connect.perform_login (self.opts['username'], self.opts['password'])

	def update_local_listing (self):

		if self.opts['localdir']:
			main.app.update_local_listing (self.opts['localdir'])

	def get_opts (self):

		# Return a copy
		opts = { }
		opts.update (self.opts)
		return opts

	def get_host (self):
		return self.opts['host']

	def get_port (self):
		return self.opts['port']

	def get_retries (self):
		return self.opts['retries']

	def set_log_active (self, bool):

		if bool:
			for key in self.logs.keys ():
				if self.logs[key] is None:
					self.logs[key] = logging.Log (os.path.join (app.CONFIG_DIR, main.app[key]), mark=0)
		else:
			for key in self.logs.keys ():
				if self.logs[key] is not None:
					old_ref = self.logs[key]
					self.logs[key] = None
					del old_ref
		self.logging = bool

	def log (self, direction, user, host, src, dest, size):

		key = direction + '_log'
		if self.logs [key] is not None:
			self.logs[key].write (_("%s@%s: %s -> %s [%d bytes]") %(user, host, src, dest, size))

	def save_transfer_state (self, remote_dir, local_dir, file, size, direction, attempt, type, method):

		self.remote_path = remote_dir
		self.local_path = local_dir
		self.filename = file
		self.size = size
		self.direction = direction
		self.attempt = attempt
		self.type = type
		self.method = method

	def initiate_transfer (self, remote_dir, local_dir, file, size, direction, attempt, method):

		self.save_transfer_state (remote_dir, local_dir, file, size, direction, attempt, 'f', method)
		self.connect.cwd (remote_dir)
		self.progress = (0, size)
		self.progressbar.update (0.0)
		self.start_time = time.time ()
		self.update_transfer_estimate ()

		# For whether or not we want to resume
		rest = None

		if self.direction == ftp.DOWNLOAD:
			local_file = os.path.join (local_dir, file)

			# Check if the file already exists and whether or not we should resume
			if os.path.exists (local_file) and os.path.isfile (local_file):
				try:
					file_size = os.path.getsize (local_file)
				except OSError, (errno, strerror):
					GtkExtra.message_box (title=_("ftpcube Error"), message=_("ftpcube Error: %s") %strerror, buttons=( [ _("Ok") ]))

				if file_size < size:
					rest = file_size
					self.progress = (0, size - file_size)

			if method == main.app.BINARY:
				self.connect.retrieve_binary (file, local_dir, self, rest=rest)
			else:
				self.connect.retrieve_ascii (file, local_dir, self)
			self.log (_('download'), self.opts['username'], self.opts['host'], string.join ([ remote_dir, file ], '/'), os.path.join (local_dir, file), size)
		elif self.direction == ftp.UPLOAD:
			local_file = os.path.join (local_dir, file)

			# Get the size of the uploading file. If we were given a size that's
			# smaller than it, then we have the size of the remote fragment and
			# we want to perform a resume
			try:
				file_size = os.path.getsize (local_file)
			except OSError, (errno, strerror):
				GtkExtra.message_box (title=_("ftpcube Error"), message=_("ftpcube Error: %s") %strerror, buttons=( [ _("Ok") ]))

			if size < file_size:
				rest = size
				self.progress = (0, file_size - size)

			if method == main.app.BINARY:
				self.connect.upload_binary (file, local_dir, self, rest=rest)
			else:
				self.connect.upload_ascii (file, local_dir, self)
			self.log (_('upload'), self.opts['username'], self.opts['host'], os.path.join (local_dir, file), string.join ([ remote_dir, file ], '/'), size)

	def recurse_directory (self, remote_dir, local_dir, file, size, direction, attempt, method):

		self.save_transfer_state (remote_dir, local_dir, file, size, direction, attempt, 'd', method)
		self.pixmap.set (self.listing_pixmap, self.listing_mask)

		if remote_dir == '/':
			remote_path = remote_dir + file
		else:
			remote_path = "%s/%s" %(remote_dir, file)

		if local_dir == '/':
			local_path = local_dir + file
		else:
			local_path = "%s/%s" %(local_dir, file)

		if direction == ftp.DOWNLOAD:
			if not os.path.exists (local_path):
				try:
					os.makedirs (local_path)
				except OSError, (errno, strerror):
					GtkExtra.message_box (title=_("ftpcube Error"), message=_("ftpcube Error: %s") %strerror, buttons=( [ _("Ok") ]))
					return

			self.connect.cwd (remote_path)

			def callback (listing, remote_path=remote_path, local_path=local_path, direction=direction, method=method, self=self):

				files = [ ]
				for file in listing:
					gtk.threads_enter ()
					entry = main.app.remotewin.parse_list_entry (file)
					gtk.threads_leave ()

					if entry is not None:
						entry = list (entry)
						# Fix up the flags
						entry[3] = entry[3][0]
						# Copy over the size into the truesize entry since
						# we don't care about mangling the size in the thread
						entry.append (entry[1])
						files.append (tuple (entry))

				gtk.threads_enter ()
				main.app.transfer (remote_path, local_path, files, direction, method)
				gtk.threads_leave ()

			# Retreive the new listing and execute the callback
			self.connect.list (callback=callback)
		else:
			self.connect.mkdir (remote_path)
			files = main.app.localwin.read_dir (local_path)
			# Purge the '..' entry
			for f in files:
				if f[0] == '..':
					files.remove (f)
			main.app.transfer (remote_path, local_path, files, direction, method)

	def recurse_link (self, remote_dir, local_dir, file, size, direction, attempt, method):

		def callback (errmsg, remote_dir=remote_dir, local_dir=local_dir, file=file, size=size, direction=direction, \
		              attempt=attempt, method=method, self=self):

			self.recurse_directory (remote_dir, local_dir, file, size, direction, attempt, method)

		# Set the callback handler in case this really wasn't really a directory
		self.connect.set_exception_handler (callback)
		self.initiate_transfer (remote_dir, local_dir, file, size, direction, attempt, method)

	def update_transfer (self, amt):

		if not self.progressbar['visible']:
			self.progress_packer.show ()
			self.progressbar.show ()
			self.pixmap.set (self.download_pixmap, self.download_mask)
		self.idle = 0

		current, total = self.progress
		self.progress = (current + amt, total)
		percentage = float (self.progress[0]) / float (self.progress[1])
		# If the percentage is great than 1.0, such as in the case of a link
		# where we don't know the real size, set it to 1.0
		if percentage > 1.0:
			percentage = 1.0
		self.progressbar.update (percentage)

	def update_transfer_estimate (self):

		# Make the necessary calculations to create the status label
		rate = self.transfer_speed ()
		if rate:
			remaining = float (self.progress[1] - self.progress[0]) / (rate * 1024.0)
		else:
			remaining = 0
		remaining = int (remaining)
		hrs = remaining / 3600
		min = (remaining % 3600) / 60
		sec = remaining % 60

		if self.direction == ftp.DOWNLOAD:
			self.label.set_text (self.get_thread_name () + _("Host: %s\nDownloading %s with %.02f kb/s, %d of %d bytes, %d:%2d:%2d sec left")
			                     %(self.opts['host'], self.filename, rate, self.progress[0],
			                       self.progress[1], hrs, min, sec))
		elif self.direction == ftp.UPLOAD:
			self.label.set_text (self.get_thread_name () + _("Host: %s\nUploading %s with %.02f kb/s, %d of %d bytes, %d:%2d:%2d sec left")
			                     %(self.opts['host'], self.filename, rate, self.progress[0],
			                       self.progress[1], hrs, min, sec))

	def isTransfering (self):

		current, total = self.progress
		if current == 0 or current == total:
			return 0
		else:
			return 1

	def transfer_speed (self):

		elapsed = time.time () - self.start_time
		if elapsed:
			return (float (self.progress[0]) / float (elapsed)) / 1024 # In KB
		else:
			return 0

	def cancel_download (self, button):

		self.label.set_text (self.get_thread_name () + _("Host: %s\nCancelling download...\n") %(self.opts['host']))
		self.pixmap.set (self.abort_pixmap, self.abort_mask)
		self.connect.abort ()

	def cancel_thread (self, button=None):

		if self.busy ():
			self.cancel_download (button)
		self.done = 1

	def abort (self):

		self.pixmap.set (self.abort_pixmap, self.abort_mask)
		self.label.set_text (self.get_thread_name () + _("Host: %s\nAborting...\n") %(self.opts['host']))
		self.idle = 0
		self.connect.abort ()

	def disable_idle_timeout (self, button):
		self.idle_disabled = 1

	def enable_idle_timeout (self, button):

		self.idle = 0
		self.idle_disabled = 0

	def reset (self):
		self.connect.reset ()

	def set_exception_handler (self, callback):
		self.connect.set_exception_handler (callback)

	def clear_exception_handler (self):
		self.connect.clear_exception_handler ()

	def finished (self):

		if not self.done:
			return self.done
		else:
			if self.connect is not None:
				self.connect.destroy ()
			return self.done

	def list (self):

		self.pixmap.set (self.listing_pixmap, self.listing_mask)
		self.label.set_text (self.get_thread_name () + _("Host: %s\nGetting Dir Listing for %s\n")
		                     %(self.opts['host'], main.app.get_remote_dir ()))
		self.idle = 0
		self.connect.list ()

	def delete (self, file):

		self.pixmap.set (self.action_pixmap, self.action_mask)
		self.label.set_text (self.get_thread_name () + _("Host: %s\nDeleting File...\n") %(self.opts['host']))
		self.idle = 0
		self.connect.delete (file)

	def rename (self, name, new_name):

		self.pixmap.set (self.action_pixmap, self.action_mask)
		self.label.set_text (self.get_thread_name () + _("Host: %s\nRenaming %s to %s...\n") %(self.opts['host'], name, new_name))
		self.idle = 0
		self.connect.rename (name, new_name)

	def mkdir (self, dir):

		self.pixmap.set (self.action_pixmap, self.action_mask)
		self.label.set_text (self.get_thread_name () + _("Host: %s\nCreating Directory...\n") %(self.opts['host']))
		self.idle = 0
		self.connect.mkdir (dir)

	def rmdir (self, file):

		self.pixmap.set (self.action_pixmap, self.action_mask)
		self.label.set_text (self.get_thread_name () + _("Host: %s\nRemoving Directory...\n") %(self.opts['host']))
		self.idle = 0
		self.connect.rmdir (file)

	def cwd (self, dir):

		self.pixmap.set (self.action_pixmap, self.action_mask)
		self.label.set_text (self.get_thread_name () + _("Host: %s\nChanging Directory...\n") %(self.opts['host']))
		self.idle = 0
		self.connect.cwd (dir)

	def update_idle (self):

		self.idle = self.idle + 1
		self.label.set_text (self.get_thread_name () + _("Host: %s\nIdle for %d secs") %(self.opts['host'], self.idle))
		self.pixmap.set (self.idle_pixmap, self.idle_mask)

	def busy (self):

		# If self.connect is none, then there was an error at some point
		# so we consider ourselves busy
		if self.connect is None:
			return 1
		else:
			return self.connect.ftp_busy ()

	def get_entry (self):
		return self.entry_box

class MainThread (FtpThread):

	def __init__ (self, connect_opts, name=None):

		FtpThread.__init__ (self, connect_opts, name=name)

		self.abort_conn = 0

	def initiate_connect (self):

		self.abort_conn = 0
		self.pixmap.set (self.connect_pixmap, self.connect_mask)
		self.label.set_text (self.get_thread_name () + _("Host: %s\nConnecting...") %(self.opts['host']))

		self.attempt = 0

		self.connect = ftp.FTP (
			host=self.opts['host'],
			port=self.opts['port'],
			silence=self.silence,
			logging=main.app['logging'],
		)

		self.connect.set_exception_handler (self.connect_exception_handler)
		self.connect.initiate_connect ()
		self.connect.set_exception_handler (self.connect_exception_handler)
		self.connect.get_welcome_msg ()
		self.connect.set_exception_handler (self.connect_exception_handler)
		self.connect.perform_login (self.opts['username'], self.opts['password'])

		# This will get wiped out of the queue if the handler is called
		self.complete_connection ()

	def complete_connection (self):

		if not self.opts['remotedir']:
			self.opts['remotedir'] = '/'
		self.connect.cwd (self.opts['remotedir'])
		main.app.update_remote_status_dir (self.opts['remotedir'])
		self.connect.list ()

	def retry_connect (self, attempt):

		self.attempt = attempt

		self.connect = ftp.FTP (
			host = self.opts['host'],
			port = self.opts['port'],
			silence=self.silence,
			logging=main.app['logging'],
		)

		self.connect.set_exception_handler (self.connect_exception_handler)
		self.connect.initiate_connect ()
		self.connect.get_welcome_msg ()
		self.connect.perform_login (self.opts['username'], self.opts['password'])

		# This will get wiped out of the queue if the handler is called
		self.complete_connection ()

	def abort (self):

		if not self.abort_conn:
			self.abort_conn = 1
		else:
			FtpThread.abort (self)

	def connect_exception_handler (self, msg):

		if self.attempt == self.opts['retries']:
			gtk.threads_enter ()
			main.app.ftp_error_tell_user (_("Maximum connection attempt hit. Aborting connection..."))
			gtk.threads_leave ()
			self.done = 1
		else:
			# Wait a bit
			delay = self.opts['delay']
			gtk.threads_enter ()
			main.app.ftp_error_tell_user (_("Waiting %d secs before re-attempting connection...") %delay)
			gtk.threads_leave ()

			slept = 0
			while slept < delay:
				time.sleep (1)
				slept = slept + 1

				# If we should abort, cancel the connection thread
				if self.abort_conn:
					self.done = 1
					break

			if self.connect is not None:
				gtk.threads_enter ()
				self.connect.destroy ()
				gtk.threads_leave ()
				self.connect = None

			if not self.abort_conn:
				self.retry_connect (self.attempt + 1)

	def initiate_transfer (self, remote_dir, local_dir, file, size, direction, attempt, method):

		old_dir = main.app.get_remote_dir ()
		FtpThread.initiate_transfer (self, remote_dir, local_dir, file, size, direction, attempt, method)
		self.connect.cwd (old_dir)

	def recurse_directory (self, remote_dir, local_dir, file, size, direction, attempt, method):

		old_dir = main.app.get_remote_dir ()
		FtpThread.recurse_directory (self, remote_dir, local_dir, file, size, direction, attempt, method)
		self.connect.cwd (old_dir)

	def recurse_link (self, remote_dir, local_dir, file, size, direction, attempt, method):

		old_dir = main.app.get_remote_dir ()
		FtpThread.recurse_link (self, remote_dir, local_dir, file, size, direction, attempt, method)
		self.connect.cwd (old_dir)

	def update_transfer (self, amt):

		# Explicitly set the pixmap every time since we may have received
		# some asynch action that might have changed it.
		self.pixmap.set (self.download_pixmap, self.download_mask)
		FtpThread.update_transfer (self, amt)

	def update_idle (self):

		FtpThread.update_idle (self)

		if self.progressbar['visible']:
			self.progressbar.hide ()
			self.progress_packer.hide ()

		# Check if we need to send a noop
		if (self.idle % self.opts['timeout']) == 0:
			self.connect.noop ()

		# Check if we should disconnect
		if self.idle == self.opts['main_idle']:
			self.cancel_thread ()

	def set_log_active (self, bool):

		self.connect.set_log_active (bool)
		FtpThread.set_log_active (self, bool)

class TransferThread (FtpThread):

	def __init__ (self, connect_opts, name=None):

		FtpThread.__init__ (self, connect_opts, name=name, silence=1)

		self.connect.set_exception_handler (self.transfer_exception_handler)

	def transfer_exception_handler (self, msg):

		gtk.threads_enter ()
		main.app.add_to_failure_list ({
			'server'      : self.opts['host'],
			'file'        : self.filename,
			'size'        : str (self.size),
			'error'       : str (msg),
			'direction'   : str (self.direction),
			'port'        : str (self.opts['port']),
			'username'    : self.opts['username'],
			'password'    : self.opts['password'],
			'remote_path' : self.remote_path,
			'local_path'  : self.local_path,
			'attempt'     : self.attempt,
			'type'        : self.type,
			'method'      : str (self.method),
		})
		gtk.threads_leave ()

	def update_idle (self):

		if not self.idle_disabled:
			FtpThread.update_idle (self)
		if self.idle >= self.opts['thread_idle']:
			self.done = 1

		if self.progressbar['visible']:
			self.progressbar.hide ()
			self.progress_packer.hide ()



#
# FtpCube
# Copyright (C) 2001 Michael Gilfix
#
# This file is part of FtpCube.
#
# You should have received a file COPYING containing license terms
# along with this program; if not, write to Michael Gilfix
# (mgilfix@eecs.tufts.edu) for a copy.
#
# This version of FtpCube is open source; you can redistribute it and/or
# modify it under the terms listed in the file COPYING.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#

import gtk, GtkExtra
import sys, os, pickle
import main, app, connectwin

from intl import _

# For compatibility with win32 pygtk
if sys.platform == "win32":
	import GTKconst
	GTK = GTKconst
else:
	import GTK

class BookmarkWin:

	"""Bookmark window object

	Contains the member functions for displaying the bookmark window and
	the functionality for managing, creating, connecting, and editing
	site bookmarks. Bookmarks are stored in the ftpcube configuration
	directory under [CONFIG_DIR]/bookmarks and are stored in pickle
	format."""

	headings = [ _('Description'), _('Host'), _('Username'), _('Directory') ]

	def __init__ (self, config_dir):
		"""Creates the bookmark object and brings up the bookmark
		window. The "bookmarks" directory is created in 'config_dir'
		if it does not already exist. The bookmark tree is then loaded
		into memory for graphical display."""

		self.window = gtk.GtkWindow ()
		self.window.set_title (_("ftpcube - Bookmarks"))
		self.window.set_usize (750, 500)
		icon_pic, icon_mask = gtk.create_pixmap_from_xpm (self.window, None, os.path.join (main.ICONS_PREFIX, "ftpcube.xpm"))
		self.window.set_icon (gtk.GtkPixmap (icon_pic, icon_mask))

		packer = gtk.GtkPacker ()
		packer.show ()
		self.window.add (packer)

		self.menu_init (packer)
		self.toolbar_init (packer)

		hbox = gtk.GtkHBox ()
		hbox.show ()
		packer.add (hbox, options=GTK.EXPAND+GTK.FILL_X+GTK.FILL_Y)
		pane = gtk.GtkHPaned ()
		pane.show ()
		hbox.add (pane)

		frame = gtk.GtkFrame ()
		frame.show ()
		pane.pack1 (frame)
		self.tree = gtk.GtkCTree ()
		self.tree.connect ("select_row", self.tree_select)
		self.tree.connect ("button_press_event", self.tree_popup_menu)
		self.tree.show ()
		frame.add (self.tree)

		frame = gtk.GtkFrame ()
		frame.show ()
		pane.pack2 (frame)
		scrolled = gtk.GtkScrolledWindow ()
		scrolled.set_policy (gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
		scrolled.show ()
		frame.add (scrolled)

		self.list = gtk.GtkCList (4, self.headings)
		self.list.connect ("select_row", self.select_items)
		self.list.connect ("button_press_event", self.list_popup_menu)
		# Set up list column defaults
		for i in range(len (self.headings)):
			self.list.set_column_width (i, 100)
			self.list.set_column_justification (i, gtk.JUSTIFY_LEFT)
		self.list.set_column_width (0, 200)
		self.list.show ()
		scrolled.add_with_viewport (self.list)

		# Compute the position of the pane divider
		pane.set_position (250)

		# Check if the bookmark directory in the config directory exists
		self.bookdir = config_dir
		path = os.path.join (self.bookdir, 'bookmarks')
		if not os.path.exists (path):
			try:
				os.mkdir (path)
				os.chmod (path, 0700)
			except OSError, strerror:
				GtkExtra.message_box (title=_("ftpcube Error"), message=_("ftpcube Error: %s") %strerror, buttons=( [ _("Ok") ]))

		self.load_pixmaps ()
		# Now load the initial bookmarks directory into the tree
		node = self.tree.insert_node (None, None, [ "bookmarks" ],
						pixmap_closed = self.folder_pic,
						mask_closed = self.folder_mask,
						pixmap_opened = self.folder_pic,
						mask_opened = self.folder_mask,
						is_leaf = gtk.FALSE)
		self.load_bookmark_tree (os.path.join (self.bookdir, "bookmarks"), parent=node)
		self.tree.expand (node)
		self.tree.select (node)
		self.display_bookmarks ()

		self.window.show ()

	def menu_init (self, container):
		"""Initializes and displays the main menu for the bookmark window."""

		ag = gtk.GtkAccelGroup ()
		item = gtk.GtkItemFactory (gtk.GtkMenuBar, "<main>", ag)
		self.window.add_accel_group (ag)

		site = self.menu_process_site
		edit = self.menu_process_edit

		item.create_items ([
			(_('/_Site'),                 None,  None,        0,   '<Branch>'),
			(_('/Site/_Connect'),         None,  site,        1,   ''),
			(_('/_Edit'),                 None,  None,        0,   '<Branch>'),
			(_('/_Edit/Add _Folder'),     None,  edit,        2,   ''),
			(_('/_Edit/Rename Folder'),   None,  edit,        3,   ''),
			(_('/_Edit/Delete Folder'),   None,  edit,        4,   ''),
			(_('/_Edit/<Separator>'),     None,  None,        0,   '<Separator>'),
			(_('/_Edit/Add _Bookmark'),   None,  edit,        5,   ''),
			(_('/_Edit/_Edit Bookmark'),  None,  edit,        6,   ''),
			(_('/_Edit/Rename Bookmark'), None,  edit,        7,   ''),
			(_('/_Edit/Delete Bookmark'), None,  edit,        8,   ''),
		])
		mainmenu = item.get_widget ('<main>')

		# Now add the menu bar into the canvas
		menubox = gtk.GtkVBox ()
		menubox.pack_start (mainmenu, expand=gtk.FALSE)
		mainmenu.show ()
		menubox.show ()
		container.add (menubox, side=gtk.SIDE_TOP, anchor=gtk.ANCHOR_NW, options=GTK.FILL_X)

	def menu_process_site (self, action, widget):
		"""Performs the appropriate 'action' for the main menu site option."""

		if action == 1:
			self.connect ()

	def menu_process_edit (self, action, widget):
		"""Performs the appropriate 'action' for the main menu edit option."""

		if action == 2:
			self.add_folder ()
		elif action == 3:
			self.rename_folder ()
		elif action == 4:
			self.remove_folder ()
		elif action == 5:
			self.add_bookmark ()
		elif action == 6:
			self.edit_bookmark ()
		elif action == 7:
			self.rename_bookmark ()
		elif action == 8:
			self.remove_bookmark ()

	def toolbar_init (self, container):
		"""Initializes and displays the bookmark toolbar."""

		toolbox = gtk.GtkHandleBox ()
		toolbar = gtk.GtkToolbar (gtk.ORIENTATION_HORIZONTAL, gtk.TOOLBAR_ICONS)
		toolbar.set_space_size (10)
		toolbar.set_space_style (GTK.TOOLBAR_SPACE_LINE)

		connect_pic, connect_mask = gtk.create_pixmap_from_xpm (self.window, None, os.path.join (main.ICONS_PREFIX, "quick_connect.xpm"))
		folder_pic, folder_mask = gtk.create_pixmap_from_xpm (self.window, None, os.path.join (main.ICONS_PREFIX, "add_folder.xpm"))
		add_pic, add_mask = gtk.create_pixmap_from_xpm (self.window, None, os.path.join (main.ICONS_PREFIX, "add_bookmark.xpm"))
		edit_pic, edit_mask = gtk.create_pixmap_from_xpm (self.window, None, os.path.join (main.ICONS_PREFIX, "edit_bookmark.xpm"))

		def connect_button (button, toolbar=toolbar, self=self):
			self.connect ()

		def folder_button (button, toolbar=toolbar, self=self):
			self.add_folder ()

		def add_button (button, toolbar=toolbar, self=self):
			self.add_bookmark ()

		def edit_button (button, toolbar=toolbar, self=self):
			self.edit_bookmark ()

		toolbar.append_item (_("Connect"), _("Connect"),
			_("Connect"),
			gtk.GtkPixmap (connect_pic, connect_mask),
			connect_button)
		toolbar.append_space ()
		toolbar.append_item (_("Add Folder"), _("Add Folder"),
			_("Add Folder"),
			gtk.GtkPixmap (folder_pic, folder_mask),
			folder_button)
		toolbar.append_item (_("Add Bookmark"), _("Add Bookmark"),
			_("Add Bookmark"),
			gtk.GtkPixmap (add_pic, add_mask),
			add_button)
		toolbar.append_item (_("Edit Bookmark"), _("Edit Bookmark"),
			_("Edit Bookmark"),
			gtk.GtkPixmap (edit_pic, edit_mask),
			edit_button)

		toolbar.show ()
		toolbox.add (toolbar)
		toolbox.show ()
		container.add (toolbox, side=gtk.SIDE_TOP, anchor=gtk.ANCHOR_NW, options=GTK.FILL_X)

	def destroy (self):
		"""Removes the bookmark window."""

		self.window.hide ()
		self.window.destroy ()

	def tree_popup_menu (self, tree, event):
		"""Displays the right-click pop-up menu when a bookmark subfolder is
		right-clicked."""

		if event.button == 3 and not tree.selection == []:
			opts = [
				(_('Rename Folder'), None, lambda b, self=self: self.rename_folder ()),
				(_('Delete Folder'), None, lambda b, self=self: self.remove_folder ()),
			]

			menu = GtkExtra.MenuFactory (GtkExtra.MENU_FACTORY_MENU)
			menu.add_entries (opts)
			menu.popup (None, None, None, event.button, event.time)
			menu.show ()

	def list_popup_menu (self, list, event):
		"""Displays the right-click pop-up menu when an individual bookmark
		is right-clicked."""

		if event.button == 3 and not self.list.selection == [ ]:
			opts = [
				(_('/Edit Bookmark'),   None, lambda b, self=self: self.edit_bookmark ()),
				(_('/Rename Bookmark'), None, lambda b, self=self: self.rename_bookmark ()),
				(_('/Delete Bookmark'), None, lambda b, self=self: self.remove_bookmark ()),
			]

			menu = GtkExtra.MenuFactory (GtkExtra.MENU_FACTORY_MENU)
			menu.add_entries (opts)
			menu.popup (None, None, None, event.button, event.time)
			menu.show ()

	def connect (self):
		"""Connect to the selected bookmark by bringing up the connect window with the
		appropriate options."""

		path = self.get_path_to_current_node ()
		name = self.list.get_text (self.list.selection.pop (0), 0)
		bookfile = os.path.join (path, name)
		options = self.load_bookmark (bookfile)

		# Check if we're breaking a connection and display the disconnect box
		# if we are
		if main.app.main_thread is None:
			connectwin.ConnectWin (main.app.initiate_connection, options)
		else:
			main.app.display_disconnect_box (main.app.main_thread.get_host (), 'ConnectWin', opts=options)

		self.destroy ()

	def select_items (self, list, r, c, event):
		"""Look for double clicks and bring up the connect window for the selected bookmark."""

		if hasattr (event, 'type'):
			# Check for double clicks
			if event.type == 5:

				self.list.selection.append (r)
				self.connect ()

	def load_pixmaps (self):

		self.folder_pic, self.folder_mask = gtk.create_pixmap_from_xpm (self.list, None, os.path.join (main.ICONS_PREFIX, "folder.xpm"))

	def tree_select (self, tree, row, col, event):
		"""Expands the selected double-clicked subfolder by loading all the bookmarks into the
		bookmark list."""

		if hasattr (event, 'type'):
			if event.type == 5:
				node = self.tree.node_nth (row)
				self.tree.expand (node)

		self.display_bookmarks ()

	def add_folder (self):
		"""A pop-up window is brough up asking for the name of the new bookmark sub-folder to
		create and the folder is then created with private permissions (if the OS supports
		permissions."""

		if not self.tree.selection:
			# Just use the top level
			node = None
			path = self.bookdir
		else:
			node = self.tree.selection.pop (0)
			path = self.get_path_to_node (node)

		folder = GtkExtra.input_box (title=_("ftpcube - Add Bookmark Folder"), message=_("Folder Name:"))

		if folder is not None:
			os.mkdir (os.path.join (path, folder), 0700)
			node = self.tree.insert_node (node, None, [ folder ],
					pixmap_closed = self.folder_pic,
					mask_closed = self.folder_mask,
					pixmap_opened = self.folder_pic,
					mask_opened = self.folder_mask,
					is_leaf = gtk.FALSE)

	def rename_folder (self):
		"""Brings up a pop-up window asking for the new folder name and then moves the
		older folder to the new name. The subfolder tree is then updated to reflect the
		new changes."""

		for node in self.tree.selection:

			# Skip if this is the root node
			if node == self.tree.base_nodes ().pop (0):
				continue

			path = self.get_path_to_node (node)
			new_folder = GtkExtra.input_box (title=_("ftpcube - Rename Folder"), message=_("Rename to:"))
			if new_folder is not None:
				new_path = os.path.join (os.path.dirname (path), new_folder)
				try:
					os.rename (path, new_path)
				except OSError, (errno, strerror):
					GtkExtra.message_box (title=_("ftpcube Error"), message=_("ftpcube Error: %s") %strerror, buttons=( [ _("Ok") ]))
				parent = node.parent
				self.tree.remove_node (node)
				node = self.tree.insert_node (parent, None, [ new_folder ],
					pixmap_closed = self.folder_pic,
					mask_closed = self.folder_mask,
					pixmap_opened = self.folder_pic,
					mask_opened = self.folder_mask,
					is_leaf = gtk.FALSE)
				self.load_bookmark_tree (new_path, parent=node)

	def remove_folder (self):
		"""Removes the selected folder and updates the tree accordingly."""

		for node in self.tree.selection:

			# Skip if this is the root node
			if node == self.tree.base_nodes ().pop (0):
				continue

			path = self.get_path_to_node (node)
			try:
				os.rmdir (path)
			except OSError, (errno, strerror):
				GtkExtra.message_box (title=_("ftpcube Error"), message=_("ftpcube Error: %s") %strerror, buttons=( [ _("Ok") ]))
				return
			self.tree.remove_node (node)

	def get_path_to_current_node (self):
		"""Recurses through the tree and returns path to the current node (Assuming an implicit
		prefix of './')."""

		# Get the current folder
		if not self.tree.selection:
			return self.get_path_to_node (self.tree.node_nth (0))
		else:
			node = self.tree.selection.pop (0)
			return self.get_path_to_node (node)

	def get_path_to_node (self, node):
		"""The recursive function to traverse the path tree."""

		if node.parent is None:
			return os.path.join (self.bookdir, (self.tree.node_get_pixtext (node, 0))[0])
		else:
			# Figure out current path
			path = (self.tree.node_get_pixtext (node, 0))[0]
			while node.parent is not None:
				node = node.parent
				path = os.path.join ((self.tree.node_get_pixtext (node, 0))[0], path)
			return os.path.join (self.bookdir, path)

	def add_bookmark (self):
		"""Adds a new bookmark by bringing up the connection window with the application
		defaults. The new bookmark is then created when the "ok" button is clicked."""

		config_opts = main.app.load_config (main.app.get_config_file ())
		connectwin.ConnectWin (self.create_bookmark, config_opts)

	def create_bookmark (self, connectwin):
		"""The callback function provided during the creation of the connection window
		object. The callback gets executed when the "ok" button is clicked and is
		provided a reference to the called window 'connectwin' for retrieval of
		the set options. The new bookmark is then created in the current bookmark
		folder."""

		opts = connectwin.get_options ()
		if opts['host']:
			bookfile = os.path.join (self.get_path_to_current_node (), opts['host'])
			self.save_bookmark (bookfile, opts)
			self.display_bookmarks ()

	def edit_bookmark (self):
		"""Brings up the connect window with the options contained in the selected bookmark
		file. Options are saved upon a click of the "ok" button."""

		path = self.get_path_to_current_node ()
		name = self.list.get_text (self.list.selection.pop (0), 0)
		bookfile = os.path.join (path, name)
		options = self.load_bookmark (bookfile)

		def after_edit (connectwin, bookfile=bookfile, self=self):

			opts = connectwin.get_options ()
			self.save_bookmark (bookfile, opts)

		connectwin.ConnectWin (after_edit, options)

	def rename_bookmark (self):
		"""Brings up a pop-up window asking for the new name of a bookmark. The selected
		bookmark is then renamed to the new name."""

		path = self.get_path_to_current_node ()
		name = self.list.get_text (self.list.selection.pop (0), 0)
		new_name = GtkExtra.input_box (title=_("ftpcube - Rename Bookmark"), message=_("Rename to:"))
		if new_name is not None:
			try:
				os.rename (os.path.join (path, name), os.path.join (path, new_name))
			except OSError, (errno, strerror):
				GtkExtra.message_box (title=_("ftpcube Error"), message=_("ftpcube Error: %s") %strerror, buttons=( [ _("Ok") ]))
			self.display_bookmarks ()

	def remove_bookmark (self):
		"""Deletes the selected bookmark and updates the bookmark listing accordingly."""

		if not self.list.selection:
			# Nothing selected so do nothing
			return

		path = self.get_path_to_current_node ()
		name = self.list.get_text (self.list.selection.pop (0), 0)
		path = os.path.join (path, name)
		try:
			os.unlink (path)
		except OSError, (errno, strerror):
			GtkExtra.message_box (title=_("ftpcube Error"), message=_("ftpcube Error: %s") %strerror, buttons=( [ _("Ok") ]))
			return
		self.display_bookmarks ()

	def save_bookmark (self, bookmark, opts):
		"""Writes out the bookmark configuration data in 'opts' to the 'bookmark' file."""

		try:
			file = open (bookmark, 'w')
		except IOError, strerror:
			GtkExtra.message_box (title=_("ftpcube Error"), message=_("ftpcube Error: %s") %strerror, buttons=( [ _("Ok") ] ))
			return
		pickler = pickle.Pickler (file)
		pickler.dump (opts)
		file.close ()

		# Set the permission of the bookmark to be sure it's private
		try:
			os.chmod (bookmark, 0600)
		except OSError, strerror:
			GtkExtra.message_box (title=_("ftpcube Error"), message=_("ftpcube Error: %s") %strerror, buttons=( [ _("Ok") ] ))

	def load_bookmark (self, file):
		"""Loads the configuration data from 'file' and returns a hash containing the data."""

		try:
			file = open (file)
		except IOError, strerror:
			GtkExtra.message_box (title=_("ftpcube Error"), message=_("ftpcube Error: %s") %strerror, buttons=( [ _("Ok") ] ))
			return
		pickler = pickle.Unpickler (file)
		bookmark = pickler.load ()
		file.close ()
		return bookmark

	def load_bookmark_tree (self, path, parent=None):
		"""Loads the recursive directory structure within the bookmark directory 'path' into
		the subfolder tree. 'parent' is used as a reference to the parent node during
		recursion through the tree."""

		try:
			files = os.listdir (path)
		except OSError, (errno, strerror):
			GtkExtra.message_box (title=_("ftpcube Error"), message=_("ftpcube Error: %s") %strerror, buttons=( [ _("Ok") ]))
			return

		if not files:
			# Then our parent is really a leaf folder
			self.tree.toggle_expansion (parent)
		else:
			# Add our children into the tree
			for file in files:
				full_path = os.path.join (path, file)
				if os.path.isdir (full_path):
					node = self.tree.insert_node (parent, None, [ file ],
						pixmap_closed = self.folder_pic,
						mask_closed = self.folder_mask,
						pixmap_opened = self.folder_pic,
						mask_opened = self.folder_mask,
						is_leaf = gtk.FALSE)
					self.load_bookmark_tree (full_path, parent=node)

	def display_bookmarks (self):
		"""Displays all of the bookmarks in the current selected bookmark folder in the
		bookmark listing window."""

		path = self.get_path_to_current_node ()

		try:
			files = os.listdir (path)
		except OSError, (errno, strerror):
			GtkExtra.message_box (title=_("ftpcube Error"), message=_("ftpcube Error: %s") %strerror, buttons=( [ _("Ok") ]))
			return

		self.list.freeze ()
		self.list.clear ()

		for file in files:
			full_path = os.path.join (path, file)
			if os.path.isfile (full_path):
				bookmark = self.load_bookmark (full_path)
				self.list.append ((file, "%s:%s" %(bookmark['host'], bookmark['port']),
				                   bookmark['username'], bookmark['remotedir']))

		self.list.thaw ()



#
# FtpCube
# Copyright (C) 2001 Michael Gilfix
#
# This file is part of FtpCube.
#
# You should have received a file COPYING containing license terms
# along with this program; if not, write to Michael Gilfix
# (mgilfix@eecs.tufts.edu) for a copy.
#
# This version of FtpCube is open source; you can redistribute it and/or
# modify it under the terms listed in the file COPYING.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#

import gtk, GtkExtra
import sys, os, pickle
import cmdwin, ctrlwin, filewin, connectwin, bookmarkwin
import optionwin, aboutwin
import main, threads, defaults

from intl import _

# For compatibility with the win32 pygtk package
if sys.platform == "win32":
	import GTKconst
	GTK = GTKconst
else:
	import GTK

if os.environ.has_key ('HOME'):
	CONFIG_DIR = os.path.join (os.environ['HOME'], ".ftpcube")
elif os.environ.has_key ('HOMEPATH'):
	CONFIG_DIR = os.path.join (os.environ['USERPROFILE'], "FtpCube Config")
else:
	CONFIG_DIR = main.PREFIX

class Application:

	"""Main application object

	The main application object sets up the framework for the main application
	window by drawing both the menu, toolbars, and skeleton boxes. Subsequent
	GUI objects are then created and given the skeleton boxes as their containers.
	The main application object also provides some interface functions that let
	various sub objects access information from other sub objects. There should
	only be one application instance, and subsequent objects expect to find the
	instance as main.app."""

	BINARY = 0
	ASCII = 1

	def __init__ (self, title):
		"""Initializes the application instance, creates the main menu and toolbar,
		and arranges the skeleton of boxes that will contain the various compound
		widgets that make up the main window."""

		# Initialize instance variables
		self.opts = { }
		self.main_thread = None # Need to set this to None explicitly

		self.config_file = self.check_configuration_files ()

		# Create the main window and set the appropriate icon
		self.main_window = gtk.GtkWindow (gtk.WINDOW_TOPLEVEL)
		self.main_window.set_title (title)
		self.main_window.connect ('destroy', self.quit)
		icon_pic, icon_mask = gtk.create_pixmap_from_xpm (self.main_window, None, os.path.join (main.ICONS_PREFIX, "ftpcube.xpm"))
		self.main_window.set_icon (gtk.GtkPixmap (icon_pic, icon_mask))

		# Load the window size preset if we have one, otherwise, default to 75% of the
		# screen size
		width = None
		height = None
		if self.opts.has_key ('win_size'):
			width, height = self['win_size']
		if width and height:
			self.set_size (width, height)
		else:
			width = int (gtk.screen_width () * 0.75)
			height = int (gtk.screen_height () * 0.75)
			self.main_window.set_default_size (width, height)

		packer = gtk.GtkPacker ()
		self.main_window.add (packer)
		packer.show ()

		self.menu_init (packer)
		self.toolbar_init (packer)

		vert_pane = gtk.GtkVPaned ()
		vert_pane.show ()
		packer.add (vert_pane, side=gtk.SIDE_TOP, anchor=gtk.ANCHOR_NW, options=GTK.PACK_EXPAND+GTK.FILL_X+GTK.FILL_Y)

		# Create the command window
		box = gtk.GtkVBox ()
		self.cmdwin = cmdwin.CmdWin (box)
		box.show ()
		vert_pane.add1 (box)

		# Create a paned thing for the fileview
		outer_paned = gtk.GtkHPaned ()
		inner_paned = gtk.GtkHPaned ()
		outer_paned.show ()
		inner_paned.show ()
		vert_pane.add2 (outer_paned)

		# Create the control window
		ctrl_box = gtk.GtkFrame ()
		ctrl_box.set_usize (250, -1)
		self.ctrlwin = ctrlwin.CtrlWin (ctrl_box)
		outer_paned.add1 (ctrl_box)
		ctrl_box.show ()
		outer_paned.add2 (inner_paned)
		# Set some parameters of the control window
		self.ctrlwin.download_tab.set_max_attempts (self.opts['retries'])
		self.ctrlwin.threads_tab.set_max_threads (3)

		# Create the local window
		# Calculate the size given the previous box so our panes look even
		box = gtk.GtkFrame ()
		self.localwin = filewin.LocalWin (box)
		inner_paned.add1 (box)
		box.show ()

		# Create the remote window
		box = gtk.GtkFrame ()
		self.remotewin = filewin.RemoteWin (box)
		inner_paned.add2 (box)
		box.show ()

		# Now set the slider accordingly
		size = (width - ctrl_box.size_request ()[0]) / 2
		inner_paned.set_position (size)

		self.main_window.show ()

	def __getitem__ (self, key):
		return self.opts[key]

	def __setitem__ (self, key, val):

		self.opts[key] = val
		return self.opts[key]

	def update (self, hash):
		self.opts.update (hash)

	def has_key (self, key):
		return self.opts.has_key (key)

	def keys (self):
		return self.opts.keys ()

	def get_opt_hash (self):

		opts = { }
		for key in self.keys ():
			opts[key] = self[key]
		return opts

	def check_configuration_files (self):
		"""Checks to see if application configuration data exists, and
		loads it if it does. If the directory containing the configuration
		information does not exist, it is created. Appropriate defaults
		are loaded if the configuration file doesn't exist."""

		config_file = os.path.join (CONFIG_DIR, "config")

		# Create configuration directory and files if necessary
		if not os.path.exists (CONFIG_DIR):
			try:
				os.mkdir (CONFIG_DIR)
				os.chmod (CONFIG_DIR, 0700)
			except OSError, strerror:
				GtkExtra.message_box (title=_("ftpcube Error"), message=_("ftpcube Error: %s") %strerror, buttons=( [ _("Ok") ]))
		
		if not os.path.exists (config_file):
			self.save_config (config_file, defaults.defaults)

		# Load the configuration file
		options = self.load_config (config_file)
		for o in options.keys ():
			self[o] = options[o]
		return config_file

	def get_config_file (self):
		return self.config_file

	def load_config (self, path):
		"""Returns a hash containing the configuration data found in
		the file at 'path'. The configuration data is a hash that is
		stored in pickle format."""

		file = open (path, 'r')
		pickler = pickle.Unpickler (file)
		config = pickler.load ()
		return config

	def save_config (self, path, opts):
		"""Saves the configuration data in the hash 'opts' into a
		pickled file found at 'path'."""

		file = open (path, 'w')
		pickler = pickle.Pickler (file)
		pickler.dump (opts)
		file.close ()

		# Set the file mode after every save just in case
		try:
			os.chmod (path, 0600)
		except OSError, strerror:
			GtkExtra.message_box (title=_("ftpcube Error"), message=_("ftpcube Error: %s") %strerror, buttons=( [ _("Ok") ]))

	def get_size (self):
		"""Returns a tuple containing the allocated width and height of the main window."""

		x, y, w, h = self.main_window.get_allocation ()
		size = (w, h)
		return size

	def set_size (self, width, height):
		"""Sets the size of the main window."""

		if width and height:
			self.main_window.set_default_size (width, height)

	def menu_init (self, container):
		"""Initializes and draws the main menu as well as sets the menu handlers."""

		# Add the keyboard accelerators to the main window
		ag = gtk.GtkAccelGroup ()
		item = gtk.GtkItemFactory (gtk.GtkMenuBar, "<main>", ag)
		self.main_window.add_accel_group (ag)

		# Assign references to menu handlers
		file = self.menu_process_file
		local = self.menu_process_local
		remote = self.menu_process_remote
		help = self.menu_process_help

		item.create_items ([
			(_('/_File'), None, None, 0, '<Branch>'),
			(_('/_File/_Connect'), None, file, 1, ''),
			(_('/_File/_Quick Connect'), None, file, 2, ''),
			(_('/_File/_Bookmarks'),  None, file, 3, ''),
			(_('/_File/_Disconnect'), None, file, 4, ''),
			(_('/_File/<Separator>'), None, None, 0, '<Separator>'),
			(_('/_File/_Options'), None, file, 5, ''),
			(_('/_File/<Separator>'), None, None, 0, '<Separator>'),
			(_('/_File/E_xit'), '<alt>F4', file, 6,''),

			(_('/_Local/_Upload Files'), None, local, 1, ''),
			(_('/_Local/_Rename Files'), None, local, 2, ''),
			(_('/_Local/_Delete Files'), None, local, 3, ''),
			(_('/_Local/<Separator>'), None, None, 0, '<Separator>'),
			(_('/_Local/_Create Directory'), None, local, 4,   ''),
			(_('/_Local/C_hange Directory'), None, local, 5, ''),

			(_('/_Remote/_Download Files'), None, remote, 1, ''),
			(_('/_Remote/_Rename Selected'), None, remote, 2, ''),
			(_('/_Remote/D_elete Files'), None, remote, 3, ''),
			(_('/_Remote/<Separator>'), None, None, 0, '<Separator>'),
			(_('/_Remote/_Create Directory'), None, remote, 4, ''),
			(_('/_Remote/Remo_ve Directories'), None, remote, 5, ''),
			(_('/_Remote/C_hange Directory'),  None, remote, 6, ''),

			(_('/_Help'), None, None, 0, '<Branch>'),
			(_('/_Help/_About'), None, help, 1, '')
		])
		mainmenu = item.get_widget ('<main>')

		# Now add the menu bar into the canvas
		menubox = gtk.GtkVBox ()
		menubox.pack_start (mainmenu, expand=gtk.FALSE)
		mainmenu.show ()
		menubox.show ()
		container.add (menubox, side=gtk.SIDE_TOP, anchor=gtk.ANCHOR_NW, options=GTK.FILL_X)

	def menu_process_file (self, action, widget):
		"""Invokes the appropriate handler for a menu option under the
		file menu."""

		if action == 0:
			pass
		elif action == 1:
			self.connect_window ()
		elif action == 2:
			self.quick_connect_window ()
		elif action == 3:
			self.bookmark_window ()
		elif action == 4:
			self.disconnect ()
		elif action == 5:
			self.options_window ()
		elif action == 6:
			self.quit (self)

	def menu_process_local (self, action, widget):
		"""Invokes the appropriate handler for a menu option under the
		file menu."""

		if action == 1:
			self.localwin.upload_selected (None, None)
		elif action == 2:
			self.localwin.rename_selected (None, None)
		elif action == 3:
			self.localwin.delete_selected (None, None)
		elif action == 4:
			self.localwin.create_directory (None, None)
		elif action == 5:
			self.localwin.cd_selected (None, None)

	def menu_process_remote (self, action, widget):
		"""Invokes the appropriate handler for a menu option under the
		remote  menu."""

		if self.main_thread:
			if action == 1:
				self.remotewin.download_selected (None, None)
			elif action == 2:
				self.remotewin.rename_selected (None, None)
			elif action == 3:
				self.remotewin.delete_selected (None, None)
			elif action == 4:
				self.remotewin.mkdir_selected (None, None)
			elif action == 5:
				self.remotewin.rmdir_selected (None, None)
			elif action == 6:
				self.remotewin.cd_selected (None, None)

	def menu_process_help (self, action, widget):
		"""Invokes the appropriate hanlder for a menu option under
		the help menu."""

		if action == 1:
			self.about_window ()

	def toolbar_init (self, container):
		"""Initializes and draws the main toolbar, as well as sets the
		button handlers."""

		toolbox = gtk.GtkHBox ()
		toolbar = gtk.GtkToolbar (gtk.ORIENTATION_HORIZONTAL, gtk.TOOLBAR_ICONS)
		toolbar.set_space_size (10)
		toolbar.set_space_style (GTK.TOOLBAR_SPACE_LINE)

		connect_pic, connect_mask = gtk.create_pixmap_from_xpm (self.main_window, None, os.path.join (main.ICONS_PREFIX, "connect.xpm"))
		quick_pic, quick_mask = gtk.create_pixmap_from_xpm (self.main_window, None, os.path.join (main.ICONS_PREFIX, "quick_connect.xpm"))
		bookmark_pic, bookmark_mask = gtk.create_pixmap_from_xpm (self.main_window, None, os.path.join (main.ICONS_PREFIX, "bookmark.xpm"))
		abort_pic, abort_mask = gtk.create_pixmap_from_xpm (self.main_window, None, os.path.join (main.ICONS_PREFIX, "abort.xpm"))
		refresh_pic, refresh_mask = gtk.create_pixmap_from_xpm (self.main_window, None, os.path.join (main.ICONS_PREFIX, "refresh.xpm"))
		disconnect_pic, disconnect_mask = gtk.create_pixmap_from_xpm (self.main_window, None, os.path.join (main.ICONS_PREFIX, "disconnect.xpm"))
		exit_pic, exit_mask = gtk.create_pixmap_from_xpm (self.main_window, None, os.path.join (main.ICONS_PREFIX, "exit.xpm"))

		def connect_button (button, toolbar=toolbar, self=self):
			self.connect_window ()

		def quick_connect_button (button, toolbar=toolbar, self=self):
			self.quick_connect_window ()

		def bookmark_button (button, toolbar=toolbar, self=self):
			self.bookmark_window ()

		def abort_button (button, toolbar=toolbar, self=self):
			self.abort ()

		def refresh_button (button, toolbar=toolbar, self=self):

			if self.main_thread is not None:
				self.remotewin.refresh_selected (None, None)

		def disconnect_button (button, toolbar=toolbar, self=self):
			self.disconnect ()

		def quit_button (button, toolbar=toolbar, self=self):
			self.quit (self.main_window)

		toolbar.append_item (_("Normal Connect"), _("Normal Connect"),
			_("Normal Connect"),
			gtk.GtkPixmap (connect_pic, connect_mask),
			connect_button)
		toolbar.append_item (_("Quick Connect"), _("Quick Connect"),
			_("Quick Connect"),
			gtk.GtkPixmap (quick_pic, quick_mask),
			quick_connect_button)
		toolbar.append_item (_("Disconnect"), _("Disconnect"),
			_("Disconnect"),
			gtk.GtkPixmap (disconnect_pic, disconnect_mask),
			disconnect_button)
		toolbar.append_item (_("Abort"), _("Abort Current Actions"),
			_("Abort Current Actions"),
			gtk.GtkPixmap (abort_pic, abort_mask),
			abort_button)
		toolbar.append_item (_("Refresh"), _("Refresh"),
			_("Refresh"),
			gtk.GtkPixmap (refresh_pic, refresh_mask),
			refresh_button)
		toolbar.append_space ()
		toolbar.append_item (_("Bookmarks"), _("Bookmarks"),
			_("Bookmarks"),
			gtk.GtkPixmap (bookmark_pic, bookmark_mask),
			bookmark_button)
		toolbar.append_space ()
		toolbar.append_item (_("Quit"), _("Quit FtpCube"),
			_("Quit FtpCube"),
			gtk.GtkPixmap (exit_pic, exit_mask),
			quit_button)

		toolbar.show ()
		toolbox.pack_start (toolbar)
		toolbox.show ()

		# Now create the transfer type buttons
		self.ascii_button = gtk.GtkRadioButton (None, _("ASCII"))
		self.ascii_button.show ()
		self.ascii_button.connect ("clicked", self.ascii_clicked)
		toolbox.pack_end (self.ascii_button, expand=GTK.FALSE)
		self.binary_button = gtk.GtkRadioButton (self.ascii_button, _("BINARY"))
		self.binary_button.show ()
		self.binary_button.connect ("clicked", self.binary_clicked)
		toolbox.pack_end (self.binary_button, expand=GTK.FALSE, padding=2)

		# Add the entry box in before the transfer type buttons
		self.dir_entry_list = [ ]
		self.dir_entry_box = gtk.GtkCombo ()
		self.dir_entry_box.entry.connect ("key_press_event", self.entry_key_pressed)
		self.dir_entry_box.list.connect ("button_press_event", self.entry_clicked)
		self.dir_entry_box.disable_activate ()
		self.dir_entry_box.show ()
		toolbox.pack_end (self.dir_entry_box, padding=5)

		if self['transfer_type'] == self.ASCII:
			self.ascii_button.set_active (gtk.TRUE)
		else:
			self.binary_button.set_active (gtk.TRUE)

		container.add (toolbox, side=gtk.SIDE_TOP, anchor=gtk.ANCHOR_NW, options=GTK.FILL_X)

	def ascii_clicked (self, button):
		"""A handler for when the ascii radio button is clicked."""

		self['transfer_type'] = self.ASCII

	def binary_clicked (self, button):
		"""A handler for when the binary radio button is clicked."""

		self['transfer_type'] = self.BINARY

	def connect_window (self):
		"""A handler for bringing up the connection window. If there is currently
		an active connection, a disconnection dialogue is displayed."""

		if self.main_thread is None:
			connectwin.ConnectWin (self.initiate_connection, self)
		else:
			self.display_disconnect_box (self.main_thread.get_host (), 'ConnectWin')

	def quick_connect_window (self):
		"""A handler for bringing up the quick connection window. If there is currently
		an active connection, a disconnection dialogue is displayed."""

		if self.main_thread is None:
			connectwin.QuickConnectWin (self.initiate_connection, self)
		else:
			self.display_disconnect_box (self.main_thread.get_host (), 'QuickConnectWin')

	def bookmark_window (self):
		"""A handler for bringing up the bookmark window"""

		bookmarkwin.BookmarkWin (CONFIG_DIR)

	def display_disconnect_box (self, host, type, opts=None):
		"""Displays a disconnection dialogue. If the user chooses to disconnect, the
		current connection is severed and the appropriate connection window displayed."""

		choice = GtkExtra.message_box (title=_("ftpcube - Disconnect?"),
		                      message=_("You are currently connected to: %s\nDo you wish to break the connection and connect to a new host?") %host,
		                      buttons=([ _("Ok"), _("Cancel") ]))

		if choice == _("Ok"):
			self.disconnect ()
			if opts is None:
				opts = self

			if type == 'ConnectWin':
				connectwin.ConnectWin (self.initiate_connection, opts)
			elif type == 'QuickConnectWin':
				connectwin.QuickConnectWin (self.initiate_connection, opts)

	def disconnect (self):
		"""Disconnects the main thread from the current active connection."""

		if self.main_thread is not None:
			self.main_thread.cancel_thread ()

		# Clear remote status stuff
		self.remotewin.clear_cache ()
		self.dir_entry_list = [ ]
		self.dir_entry_box.entry.set_text ('')

	def options_window (self):
		"""A handler for bringing up the option window."""

		optionwin.OptionWindow ()

	def about_window (self):
		"""A handler for bringing up the about window."""

		aboutwin.AboutWindow ()

	def initiate_connection (self, connectwin):
		"""Extracts the connection information (a hash) from the 'connectwin'
		object and creates a main thread object with the connection information."""

		opts = { }
		# Seed the options with the main ones
		opts.update (self.opts)
		# Update the defaults to reflect the connect window
		opts.update (connectwin.get_options ())

		self.main_thread = threads.MainThread (opts, name=_("MainThread"))
		self.ctrlwin.attach_main_thread (self.main_thread)

	def abort (self):
		"""Attempts to abort the current action of the main thread."""

		if self.main_thread is not None:
			self.main_thread.abort ()

	def display_banner (self):
		"""Prints the initial greeting in the command window."""

		self.cmdwin.display_banner ()

	def ftp_local_tell_user (self, msg):
		self.cmdwin.insert_color_text ('local', msg)

	def ftp_remote_tell_user (self, msg):
		self.cmdwin.insert_color_text ('remote', msg)

	def ftp_error_tell_user (self, msg):
		self.cmdwin.insert_color_text ('error', msg)

	def update_remote_status_dir (self, dir):
		self.remotewin.update_status_dir (dir)

	def entry_key_pressed (self, entry, event):

		# Update with an enter
		if self.main_thread and event.string == '\r':
			self.entry_action ()

	def entry_clicked (self, button, event):

		if self.main_thread:
			self.entry_action ()

	def entry_action (self):

		dir = self.dir_entry_box.entry.get_text ()
		self.main_thread.cwd (dir)
		self.remotewin.update_status_dir (dir)
		self.remotewin.get_listing ()

	def update_dir_entry_status (self, dir):

		if not dir in self.dir_entry_list:
			self.dir_entry_list.append (dir)

		self.dir_entry_box.set_popdown_strings (self.dir_entry_list)
		self.dir_entry_box.entry.set_text (dir)

	def get_ftp_listing (self):
		self.main_thread.list ()

	def update_remote_listing (self, dir):
		self.remotewin.update_listing (dir)

	def update_local_listing (self, dir):
		self.localwin.update_listing (dir)

	def get_local_dir (self):
		return self.localwin.get_dir ()

	def get_remote_dir (self):
		return self.remotewin.get_dir ()

	def update_list_transfer (self, amt):
		"""Updates the status bar in the remote file window to display the
		number of bytes currently transferred."""

		remote_dir = self.get_remote_dir ()
		self.remotewin.update_status (_("Getting listing for %s: %d bytes") %(remote_dir, amt))

	def list_transfer_finished (self):

		self.remotewin.update_status_dir (self.remotewin.get_dir ())

	def transfer (self, remote_path, local_path, files, direction, method):
		"""Adds the list of files into the transfer queue via the control
		window object."""

		main_opts = self.main_thread.get_opts ()

		self.ctrlwin.add_to_queue ({
			'files'       : files,
			'server'      : main_opts['host'],
			'port'        : main_opts['port'],
			'username'    : main_opts['username'],
			'password'    : main_opts['password'],
			'remote_path' : remote_path,
			'local_path'  : local_path,
			'maxattempts' : main_opts['retries'],
			'direction'   : direction,
			'method'      : method
		})

	def get_thread_list (self):
		return self.ctrlwin.get_thread_list ()

	def add_to_download_list (self, file):
		self.ctrlwin.add_to_downloads (file)

	def add_to_failure_list (self, hash):
		self.ctrlwin.add_to_failure (hash)

	def get_cwd (self):
		return self.localwin.get_dir ()

	def get_file_size (self, file):
		return self.remotewin.get_file_size (file)

	def quit (self, win):

		# Save the size before exiting
		self['win_size'] = self.get_size ()
		config_file = os.path.join (CONFIG_DIR, "config")
		self.save_config (config_file, self.get_opt_hash ())

		gtk.mainquit ()

# Conversion function from bytes sizes to pretty strings
def beautify_size (size):
	"""Returns a string describing the number of bytes 'size'
	in Mneumonic form."""

	if size / 1073741824:
		return _("%0.2f Gbytes") %(float (size) / 1073741824.0)
	elif size / 1048576:
		return _("%0.2f Mbytes") %(float (size) / 1048576.0)
	elif size / 1024:
		return _("%0.2f KBytes") %(float (size) / 1024.0)
	else:
		return _("%s Bytes") %(size)
	return None



#
# FtpCube
# Copyright (C) 2001 Michael Gilfix
#
# This file is part of FtpCube.
#
# You should have received a file COPYING containing license terms
# along with this program; if not, write to Michael Gilfix
# (mgilfix@eecs.tufts.edu) for a copy.
#
# This version of FtpCube is open source; you can redistribute it and/or
# modify it under the terms listed in the file COPYING.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#

defaults = {
	'host'          : None,
	'port'          : 21,
	'username'      : 'anonymous',
	'password'      : 'guest@anon.com',
	'timeout'       : 120,
	'localdir'      : None,
	'limit'         : None,
	'thread_idle'   : 30,
	'main_idle'     : 300,
	'delay'         : 5,
	'retries'       : 5,
	'proxy'         : 0,
	'remotedir'     : None,
	'logging'       : 0,
	'cmd_log'       : 'commands.txt',
	'download_log'  : 'download.txt',
	'upload_log'    : 'upload.txt',
	'local_color'   : (0.21482819734186889, 0.09512286143262505, 0.43703703703703706),
	'remote_color'  : (0.16822981105023022, 0.37037037037037035, 0.18973221780044075),
	'error_color'   : (0.45185185185185184, 0.092204513827356852, 0.15133532313537504),
	'transfer_type' : 0,
	'win_size'      : (0, 0)
}



#
# FtpCube
# Copyright (C) 2001 Michael Gilfix
#
# This file is part of FtpCube.
#
# You should have received a file COPYING containing license terms
# along with this program; if not, write to Michael Gilfix
# (mgilfix@eecs.tufts.edu) for a copy.
#
# This version of FtpCube is open source; you can redistribute it and/or
# modify it under the terms listed in the file COPYING.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#

import sys, os

from intl import _

# Try to figure out where the icon directory might be.
# For now we look in a few simple places
if sys.platform == "win32":
	PREFIX = "."
elif os.name == "posix":
	if os.path.exists ("/usr/local/share/ftpcube"):
		PREFIX = "/usr/local/share/ftpcube"
	elif os.path.exists ("/usr/share/ftpcube"):
		PREFIX = "/usr/share/ftpcube"
	else:
		PREFIX = "."
else:
	print _("ftpcube can't determine the location of it's icons on your OS.")
	print _("Please set the prefix manually in main.py")
	sys.exit (1)

import getopt, gtk
import app

ICONS_PREFIX = os.path.join (PREFIX, "icons")

cmdline_opts = { }

def main ():
	"""Start up procedure for the application. This function performs
	all necessary argument parsing and creates the application instance."""

	try:
		opts, args = getopt.getopt (sys.argv[1:], 'h', [
			'help',         # Print out help usage info
			'debug-level=', # The debugging output level to use
			'icons=',       # Directory where the icons are stored
		])
	except getopt.GetoptError:
		usage ()
		sys.exit (1)

	for o, a in opts:
		if o in ("-h", "--help"):
			usage ()
			sys.exit (1)
		elif o == "--debug-level":
			cmdline_opts['debug_level'] = a
		elif o == "--icons":
			# Override the icon finding code
			global PREFIX, ICONS_PREFIX
			PREFIX = os.path.dirname (a)
			ICONS_PREFIX = a

	# Instantiate the application and enter the main loop
	create_global_app (app.Application (_("ftpcube")))
	app.display_banner ()

	gtk.threads_enter ()
	gtk.mainloop ()
	gtk.threads_leave ()

def create_global_app (a):
	"""A convenience function for moving the application instance a
	into global namespace while avoiding any name collision."""

	global app
	app = a

def usage ():
	"""Prints a command-line usage string."""

	pass
